#ifndef CELL_BY_CELL_LIMITATION_HPP
#define CELL_BY_CELL_LIMITATION_HPP

#include <mesh/Mesh.hpp>
#include <mesh/MeshData.hpp>

#include <language/utils/InterpolateItemArray.hpp>
#include <mesh/ItemValueUtils.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshData.hpp>
#include <mesh/MeshDataManager.hpp>
#include <mesh/MeshEdgeBoundary.hpp>
#include <mesh/MeshFaceBoundary.hpp>
#include <mesh/MeshFlatEdgeBoundary.hpp>
#include <mesh/MeshFlatFaceBoundary.hpp>
#include <mesh/MeshFlatNodeBoundary.hpp>
#include <mesh/MeshNodeBoundary.hpp>
#include <mesh/MeshTraits.hpp>
#include <mesh/MeshVariant.hpp>
#include <mesh/StencilManager.hpp>
#include <scheme/DiscreteFunctionDPk.hpp>
#include <scheme/DiscreteFunctionDPkVariant.hpp>
#include <scheme/DiscreteFunctionDPkVector.hpp>
#include <scheme/DiscreteFunctionP0.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/InflowListBoundaryConditionDescriptor.hpp>
#include <scheme/PolynomialReconstruction.hpp>
#include <scheme/PolynomialReconstructionDescriptor.hpp>
#include <utils/PugsTraits.hpp>
#include <variant>

template <MeshConcept MeshTypeT>
class CellByCellLimitation
{
 private:
  using MeshType                    = MeshTypeT;
  static constexpr size_t Dimension = MeshType::Dimension;

  using Rdxd = TinyMatrix<Dimension>;
  using Rd   = TinyVector<Dimension>;

 public:
  void
  density_limiter(const MeshType& mesh,
                  const std::vector<std::shared_ptr<const IBoundaryDescriptor>>& symmetry_boundary_descriptor_list,
                  const DiscreteFunctionP0<const double>& rho,
                  DiscreteFunctionDPk<Dimension, double>& rho_L) const
  {
    const auto& cell_to_face_matrix = mesh.connectivity().cellToFaceMatrix();
    const auto& face_to_node_matrix = mesh.connectivity().faceToNodeMatrix();

    const auto& xr = mesh.xr();
    // const auto& xl = mesh.xl();

    MeshData<MeshType>& mesh_data = MeshDataManager::instance().getMeshData(mesh);

    //    auto stencil = StencilManager::instance()._getStencilArray(mesh.connectivity(), 1);
    auto stencil = StencilManager::instance()
                     .getCellToCellStencilArray(mesh.connectivity(),
                                                StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes},
                                                symmetry_boundary_descriptor_list);
    auto xj        = mesh_data.xj();
    const auto& xl = mesh_data.xl();

    // const QuadratureFormula<1> qf =
    //   QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(m_quadrature_degree));

    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
        const double rhoj = rho[cell_id];

        double rho_min = rhoj;
        double rho_max = rhoj;

        const auto cell_stencil = stencil[cell_id];
        for (size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell) {
          rho_min = std::min(rho_min, rho[cell_stencil[i_cell]]);
          rho_max = std::max(rho_max, rho[cell_stencil[i_cell]]);
        }

        double rho_bar_min = rhoj;
        double rho_bar_max = rhoj;

        for (size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell) {
          const CellId cell_k_id = cell_stencil[i_cell];
          const double rho_xk    = rho_L[cell_id](xj[cell_k_id]);

          rho_bar_min = std::min(rho_bar_min, rho_xk);
          rho_bar_max = std::max(rho_bar_max, rho_xk);
        }

        if constexpr (Dimension == 2) {
          auto face_list = cell_to_face_matrix[cell_id];

          for (size_t i_face = 0; i_face < face_list.size(); ++i_face) {
            const FaceId face_id = face_list[i_face];

            const Rd& x0 = xr[face_to_node_matrix[face_id][0]];
            const Rd& x2 = xr[face_to_node_matrix[face_id][1]];
            const Rd& x1 = .5 * (x0 + x2);

            const double rho_x0 = rho_L[cell_id](x0);
            const double rho_x1 = rho_L[cell_id](x1);
            const double rho_x2 = rho_L[cell_id](x2);

            rho_bar_min = std::min(rho_bar_min, std::min(rho_x0, std::min(rho_x1, rho_x2)));
            rho_bar_max = std::max(rho_bar_max, std::max(rho_x0, std::max(rho_x1, rho_x2)));
          }
          // const LineParabolicTransformation<Dimension> t(x0, x1, x2);
          // for (size_t iq = 0; iq < qf.numberOfPoints(); ++iq) {
          // const double rho_xk = rho_L[cell_id](t(qf.point(iq)));

          // rho_bar_min = std::min(rho_bar_min, rho_xk);
          // rho_bar_max = std::max(rho_bar_max, rho_xk);
          //}
        } else {
        }

        const double eps = 1E-14;
        double coef1     = 1;
        if (std::abs(rho_bar_max - rhoj) > eps) {
          coef1 = (rho_max - rhoj) / ((rho_bar_max - rhoj));
        }

        double coef2 = 1.;
        if (std::abs(rho_bar_min - rhoj) > eps) {
          coef2 = (rho_min - rhoj) / ((rho_bar_min - rhoj));
        }

        const double lambda = std::max(0., std::min(1., std::min(coef1, coef2)));

        auto coefficients = rho_L.coefficients(cell_id);

        coefficients[0] = (1 - lambda) * rho[cell_id] + lambda * coefficients[0];

        for (size_t i = 1; i < coefficients.size(); ++i) {
          coefficients[i] *= lambda;
        }
      });
  }

  void
  specific_internal_nrj_limiter(
    const MeshType& mesh,
    const std::vector<std::shared_ptr<const IBoundaryDescriptor>>& symmetry_boundary_descriptor_list,

    // const DiscreteFunctionP0<const double>& rho,
    // const DiscreteFunctionDPk<Dimension, double>& rho_L,
    const DiscreteFunctionP0<const double>& epsilon,
    // const DiscreteFunctionDPk<Dimension, double>&
    auto epsilon_R(const CellId cell_id, const Rd& x),
    CellValue<double>& lambda_epsilon)   // const
  //                                DiscreteFunctionDPk<Dimension, double>& epsilon_R) const
  {
    const auto& cell_to_face_matrix = mesh.connectivity().cellToFaceMatrix();
    const auto& face_to_node_matrix = mesh.connectivity().faceToNodeMatrix();

    const auto& xr = mesh.xr();
    const auto& xl = mesh.xl();

    MeshData<MeshType>& mesh_data = MeshDataManager::instance().getMeshData(mesh);

    //    auto stencil = StencilManager::instance()._getStencilArray(mesh.connectivity(), 1);
    auto stencil = StencilManager::instance()
                     .getCellToCellStencilArray(mesh.connectivity(),
                                                StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes},
                                                symmetry_boundary_descriptor_list);
    auto xj = mesh_data.xj();

    //    const QuadratureFormula<1> qf =
    // QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor(m_quadrature_degree));

    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
        const double epsilonj = epsilon[cell_id];

        double epsilon_min = epsilonj;
        double epsilon_max = epsilonj;

        const auto cell_stencil = stencil[cell_id];
        for (size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell) {
          epsilon_min = std::min(epsilon_min, epsilon[cell_stencil[i_cell]]);
          epsilon_max = std::max(epsilon_max, epsilon[cell_stencil[i_cell]]);
        }

        double epsilon_R_min = epsilonj;
        double epsilon_R_max = epsilonj;

        for (size_t i_cell = 0; i_cell < cell_stencil.size(); ++i_cell) {
          const CellId cell_k_id  = cell_stencil[i_cell];
          const double epsilon_xk = epsilon_R(cell_id, xj[cell_k_id]);

          epsilon_R_min = std::min(epsilon_R_min, epsilon_xk);
          epsilon_R_max = std::max(epsilon_R_max, epsilon_xk);
        }

        auto face_list = cell_to_face_matrix[cell_id];
        for (size_t i_face = 0; i_face < face_list.size(); ++i_face) {
          const FaceId face_id = face_list[i_face];

          const Rd& x0            = xr[face_to_node_matrix[face_id][0]];
          const Rd& x1            = xl[face_id][0];
          const Rd& x2            = xr[face_to_node_matrix[face_id][1]];
          const double epsilon_x0 = epsilon_R(cell_id, x0);
          const double epsilon_x1 = epsilon_R(cell_id, x1);
          const double epsilon_x2 = epsilon_R(cell_id, x2);

          epsilon_R_min = std::min(epsilon_R_min, std::min(epsilon_x0, std::min(epsilon_x1, epsilon_x2)));
          epsilon_R_max = std::max(epsilon_R_max, std::max(epsilon_x0, std::max(epsilon_x1, epsilon_x2)));

          // const LineParabolicTransformation<Dimension> t(x0, x1, x2);
          // for (size_t iq = 0; iq < qf.numberOfPoints(); ++iq) {
          // const double epsilon_xk = epsilon_R(cell_id, t(qf.point(iq)));

          // epsilon_R_min = std::min(epsilon_R_min, epsilon_xk);
          // epsilon_R_max = std::max(epsilon_R_max, epsilon_xk);
          //}
        }

        const double eps = 1E-14;
        double coef1     = 1;
        if (std::abs(epsilon_R_max - epsilonj) > eps) {
          coef1 = (epsilon_max - epsilonj) / ((epsilon_R_max - epsilonj));
        }

        double coef2 = 1.;
        if (std::abs(epsilon_R_min - epsilonj) > eps) {
          coef2 = (epsilon_min - epsilonj) / ((epsilon_R_min - epsilonj));
        }

        lambda_epsilon[cell_id] = std::max(0., std::min(1., std::min(coef1, coef2)));
      });
  }

  void
  computeLimitorVolumicScalarQuantityMinModDukowiczGradient(const MeshType& mesh,
                                                            const CellValue<double>& q,
                                                            const CellValue<Rd>& gradq,
                                                            CellValue<double>& Limitor_q,
                                                            const bool enableWeakBoundPositivityOnly = false) const
  {
    Assert(Dimension == 2 or Dimension == 3);
    const auto& cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();
    const auto& cell_to_face_matrix = mesh.connectivity().cellToFaceMatrix();
    const auto& node_to_cell_matrix = mesh.connectivity().nodeToCellMatrix();

    const auto& xr = mesh.xr();

    MeshData<MeshType>& mesh_data = MeshDataManager::instance().getMeshData(mesh);

    // auto volumes_cell = mesh_data.Vj();

    const auto& xcentroid = mesh_data.xj();
    const auto& xface     = mesh_data.xl();

    if (Dimension == 2) {
      parallel_for(
        mesh.numberOfCells(),
        PUGS_LAMBDA(CellId j) {   // rho_bar_lim[j] = rho[j] + Limitor_rho[j] * (rho_bar[j] - rho[j]); });
          const double State = q[j];
          const Rd& Cent     = xcentroid[j];   // mesh.xi(j);
          const Rd& Gradl    = gradq[j];

          // Min et Max des valeurs voisines

          double minvalvois(State);
          double maxvalvois(State);

          // Calcul des extremas de la fonction

          double minval(std::numeric_limits<double>::max());
          double maxval(-std::numeric_limits<double>::max());

          // At nodes
          const auto& cell_to_node = cell_to_node_matrix[j];
          // At faces
          const auto& cell_to_face = cell_to_face_matrix[j];

          for (size_t l = 0; l < cell_to_node.size(); ++l) {
            const NodeId& node       = cell_to_node[l];
            const auto& node_to_cell = node_to_cell_matrix[node];

            for (size_t k = 0; k < node_to_cell.size(); ++k) {
              const CellId& cellvois     = node_to_cell[k];
              const double statevoiscell = q[cellvois];
              const Rd Centvois          = xcentroid[cellvois];   // mesh.element()[mayvois].centroid().x();
              const Rd Mvois(Centvois - Cent);

              const double valVois = State + dot(Gradl, Mvois);   // Partie MinMod

              minval = std::min(minval, valVois);
              maxval = std::max(maxval, valVois);

              minvalvois = std::min(minvalvois, statevoiscell);
              maxvalvois = std::max(maxvalvois, statevoiscell);
            }

            const Rd M(xr[node] - Cent);

            const double valLineO2node = State + dot(Gradl, M);   // Partie Dukowicz

            minval = std::min(minval, valLineO2node);
            maxval = std::max(maxval, valLineO2node);
          }

          for (size_t l = 0; l < cell_to_face.size(); ++l) {
            const FaceId& face = cell_to_face[l];

            const Rd Mb(xface[face] - Cent);

            const double valLineO2bras = State + dot(Gradl, Mb);   // Idem aux demi faces

            minval = std::min(minval, valLineO2bras);
            maxval = std::max(maxval, valLineO2bras);
          }

          static const double epsZer0 = 1e-15;
          // on applique le traitement idem de l'ordre 2..
          double coef1 = 0, coef2 = 0;
          if (enableWeakBoundPositivityOnly) {
            coef1                  = 1;
            const double minGlobal = 1e-12;

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minGlobal - State) / ((minval - State));

          } else {
            if (std::fabs(maxval - State) <= epsZer0)   // epsIsEqualAbs())
              coef1 = 1.;
            else
              coef1 = (maxvalvois - State) / ((maxval - State));

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minvalvois - State) / ((minval - State));
          }
          const double alfa = std::max(0., std::min(1., std::min(coef1, coef2)));

          Limitor_q[j] = alfa;
          // grad(i) *= alfa;
        });
    } else {
      // throw NormalError(" Limiteur Scal Grad non encore implemente 3D");
      const auto& cell_to_edge_matrix = mesh.connectivity().cellToEdgeMatrix();
      const auto& xedge               = mesh_data.xe();

      parallel_for(
        mesh.numberOfCells(),
        PUGS_LAMBDA(CellId j) {   // rho_bar_lim[j] = rho[j] + Limitor_rho[j] * (rho_bar[j] - rho[j]); });
          const double State = q[j];
          const Rd& Cent     = xcentroid[j];   // mesh.xi(j);
          const Rd& Gradl    = gradq[j];

          // Min et Max des valeurs voisines

          double minvalvois(State);
          double maxvalvois(State);

          // Calcul des extremas de la fonction

          double minval(std::numeric_limits<double>::max());
          double maxval(-std::numeric_limits<double>::max());

          // At nodes
          const auto& cell_to_node = cell_to_node_matrix[j];
          // At faces
          const auto& cell_to_face = cell_to_face_matrix[j];
          // At faces
          const auto& cell_to_edge = cell_to_edge_matrix[j];

          for (size_t l = 0; l < cell_to_node.size(); ++l) {
            const NodeId& node       = cell_to_node[l];
            const auto& node_to_cell = node_to_cell_matrix[node];

            for (size_t k = 0; k < node_to_cell.size(); ++k) {
              const CellId& cellvois     = node_to_cell[k];
              const double statevoiscell = q[cellvois];
              const Rd Centvois          = xcentroid[cellvois];   // mesh.element()[mayvois].centroid().x();
              const Rd Mvois(Centvois - Cent);

              const double valVois = State + dot(Gradl, Mvois);   // Partie MinMod

              minval = std::min(minval, valVois);
              maxval = std::max(maxval, valVois);

              minvalvois = std::min(minvalvois, statevoiscell);
              maxvalvois = std::max(maxvalvois, statevoiscell);
            }

            const Rd M(xr[node] - Cent);

            const double valLineO2node = State + dot(Gradl, M);   // Partie Dukowicz

            minval = std::min(minval, valLineO2node);
            maxval = std::max(maxval, valLineO2node);
          }

          for (size_t l = 0; l < cell_to_face.size(); ++l) {
            const FaceId& face = cell_to_face[l];

            const Rd Mb(xface[face] - Cent);

            const double valLineO2face = State + dot(Gradl, Mb);   // Idem aux demi faces

            minval = std::min(minval, valLineO2face);
            maxval = std::max(maxval, valLineO2face);
          }

          for (size_t l = 0; l < cell_to_edge.size(); ++l) {
            const EdgeId& edge = cell_to_edge[l];

            const Rd Me(xedge[edge] - Cent);

            const double valLineO2edge = State + dot(Gradl, Me);   // Idem aux demi faces

            minval = std::min(minval, valLineO2edge);
            maxval = std::max(maxval, valLineO2edge);
          }

          static const double epsZer0 = 1e-15;
          // on applique le traitement idem de l'ordre 2..
          double coef1 = 0, coef2 = 0;
          if (enableWeakBoundPositivityOnly) {
            coef1                  = 1;
            const double minGlobal = 1e-12;

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minGlobal - State) / ((minval - State));

          } else {
            if (std::fabs(maxval - State) <= epsZer0)   // epsIsEqualAbs())
              coef1 = 1.;
            else
              coef1 = (maxvalvois - State) / ((maxval - State));

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minvalvois - State) / ((minval - State));
          }
          const double alfa = std::max(0., std::min(1., std::min(coef1, coef2)));

          Limitor_q[j] = alfa;
          // grad(i) *= alfa;
        });
    }
    synchronize(Limitor_q);
  }

  void
  computeLimitorInternalNrjScalarQuantityMinModDukowiczGradient(const MeshType& mesh,
                                                                const CellValue<double>& rho,
                                                                const CellValue<Rd>& gradrho,
                                                                const CellValue<double>& internalnrj,
                                                                const CellValue<Rd>& gradinternalnrj,
                                                                // const CellValue<Rd>& u,
                                                                const CellValue<Rdxd>& gradu,
                                                                CellValue<double>& Limitor_eps,
                                                                const bool enableWeakBoundPositivityOnly = false) const
  {
    Assert(Dimension == 2 or Dimension == 3);
    const auto& cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();
    const auto& cell_to_face_matrix = mesh.connectivity().cellToFaceMatrix();
    const auto& node_to_cell_matrix = mesh.connectivity().nodeToCellMatrix();
    const auto& xr                  = mesh.xr();

    MeshData<MeshType>& mesh_data = MeshDataManager::instance().getMeshData(mesh);

    const auto& xcentroid = mesh_data.xj();
    const auto& xface     = mesh_data.xl();

    if (Dimension == 2) {
      parallel_for(
        mesh.numberOfCells(),
        PUGS_LAMBDA(CellId j) {   // rho_bar_lim[j] = rho[j] + Limitor_rho[j] * (rho_bar[j] - rho[j]); });
          const double rhol   = rho[j];
          const Rd& Cent      = xcentroid[j];
          const Rd& grad_rhol = gradrho[j];

          // const Rd& ul        = u[j];
          const Rdxd& grad_ul = gradu[j];

          const double internal_nrj = internalnrj[j];
          // const double kinetic_nrj  = .5 * dot(ul, ul);

          const Rd& grad_internal_nrj = gradinternalnrj[j];

          // const Rd& grad_kinetic_internal_nrj = gradeps[j];

          // const double qtt_limitor_density = rhol / (rhol + dot(grad,(x-xc));
          // At nodes
          const auto& cell_to_node = cell_to_node_matrix[j];
          // At faces
          const auto& cell_to_face = cell_to_face_matrix[j];

          // Min et Max des valeurs voisines
          const double State = internal_nrj;

          double minvalvois(State);
          double maxvalvois(State);

          // Calcul des extremas de la fonction

          double minval(std::numeric_limits<double>::max());
          double maxval(-std::numeric_limits<double>::max());

          for (size_t l = 0; l < cell_to_node.size(); ++l) {
            const NodeId& node       = cell_to_node[l];
            const auto& node_to_cell = node_to_cell_matrix[node];

            for (size_t k = 0; k < node_to_cell.size(); ++k) {
              const CellId& cellvois = node_to_cell[k];
              const Rd Centvois      = xcentroid[cellvois];
              const Rd Mvois(Centvois - Cent);

              const double qtt_limitor_density = rhol / (rhol + dot(grad_rhol, Mvois));

              const Rd& High_ordervelocity       = qtt_limitor_density * (grad_ul * Mvois);
              const double High_orderinternalnrj = qtt_limitor_density * dot(grad_internal_nrj, Mvois);

              const double valVois = internal_nrj + High_orderinternalnrj -
                                     .5 * dot(High_ordervelocity, High_ordervelocity);   // Partie MinMod

              minval = std::min(minval, valVois);
              maxval = std::max(maxval, valVois);

              const double statevoiscell = internalnrj[cellvois];

              minvalvois = std::min(minvalvois, statevoiscell);
              maxvalvois = std::max(maxvalvois, statevoiscell);
            }

            const Rd Mr(xr[node] - Cent);

            const double qtt_limitor_density = rhol / (rhol + dot(grad_rhol, Mr));

            const Rd& High_ordervelocity       = qtt_limitor_density * (grad_ul * Mr);
            const double High_orderinternalnrj = qtt_limitor_density * dot(grad_internal_nrj, Mr);

            const double valLineO2node = internal_nrj + High_orderinternalnrj -
                                         .5 * dot(High_ordervelocity, High_ordervelocity);   // Partie Dukowicz

            minval = std::min(minval, valLineO2node);
            maxval = std::max(maxval, valLineO2node);
          }

          for (size_t l = 0; l < cell_to_face.size(); ++l) {
            const FaceId& face = cell_to_face[l];

            const Rd Mb(xface[face] - Cent);

            const double qtt_limitor_density = rhol / (rhol + dot(grad_rhol, Mb));

            const Rd& High_ordervelocity       = qtt_limitor_density * (grad_ul * Mb);
            const double High_orderinternalnrj = qtt_limitor_density * dot(grad_internal_nrj, Mb);

            const double valLineO2face = internal_nrj + High_orderinternalnrj -
                                         .5 * dot(High_ordervelocity, High_ordervelocity);   // Partie Dukowicz

            minval = std::min(minval, valLineO2face);
            maxval = std::max(maxval, valLineO2face);
          }

          static const double epsZer0 = 1e-15;

          // on applique le traitement idem de l'ordre 2..
          double coef1 = 0, coef2 = 0;
          if (enableWeakBoundPositivityOnly) {
            coef1                  = 1;
            const double minGlobal = 1e-12;

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minGlobal - State) / ((minval - State));

          } else {
            if (std::fabs(maxval - State) <= epsZer0)   // epsIsEqualAbs())
              coef1 = 1.;
            else
              coef1 = (maxvalvois - State) / ((maxval - State));

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minvalvois - State) / ((minval - State));
          }
          const double alfa = std::max(0., std::min(1., std::min(coef1, coef2)));

          Limitor_eps[j] = alfa;
        });
    } else {
      // throw NormalError(" Limiteur Scal Grad internal nrj non encore implemente 3D");
      const auto& cell_to_edge_matrix = mesh.connectivity().cellToEdgeMatrix();

      const auto& xedge = mesh_data.xe();

      parallel_for(
        mesh.numberOfCells(),
        PUGS_LAMBDA(CellId j) {   // rho_bar_lim[j] = rho[j] + Limitor_rho[j] * (rho_bar[j] - rho[j]); });
          const double rhol   = rho[j];
          const Rd& Cent      = xcentroid[j];
          const Rd& grad_rhol = gradrho[j];

          // const Rd& ul        = u[j];
          const Rdxd& grad_ul = gradu[j];

          const double internal_nrj = internalnrj[j];
          // const double kinetic_nrj  = .5 * dot(ul, ul);

          const Rd& grad_internal_nrj = gradinternalnrj[j];

          // const Rd& grad_kinetic_internal_nrj = gradeps[j];

          // const double qtt_limitor_density = rhol / (rhol + dot(grad,(x-xc));
          // At nodes
          const auto& cell_to_node = cell_to_node_matrix[j];
          // At faces
          const auto& cell_to_face = cell_to_face_matrix[j];
          // At edges
          const auto& cell_to_edge = cell_to_edge_matrix[j];

          // Min et Max des valeurs voisines
          const double State = internal_nrj;

          double minvalvois(State);
          double maxvalvois(State);

          // Calcul des extremas de la fonction

          double minval(std::numeric_limits<double>::max());
          double maxval(-std::numeric_limits<double>::max());

          for (size_t l = 0; l < cell_to_node.size(); ++l) {
            const NodeId& node       = cell_to_node[l];
            const auto& node_to_cell = node_to_cell_matrix[node];

            for (size_t k = 0; k < node_to_cell.size(); ++k) {
              const CellId& cellvois = node_to_cell[k];
              const Rd Centvois      = xcentroid[cellvois];
              const Rd Mvois(Centvois - Cent);

              const double qtt_limitor_density = rhol / (rhol + dot(grad_rhol, Mvois));

              const Rd& High_ordervelocity       = qtt_limitor_density * (grad_ul * Mvois);
              const double High_orderinternalnrj = qtt_limitor_density * dot(grad_internal_nrj, Mvois);

              const double valVois = internal_nrj + High_orderinternalnrj -
                                     .5 * dot(High_ordervelocity, High_ordervelocity);   // Partie MinMod

              minval = std::min(minval, valVois);
              maxval = std::max(maxval, valVois);

              const double statevoiscell = internalnrj[cellvois];

              minvalvois = std::min(minvalvois, statevoiscell);
              maxvalvois = std::max(maxvalvois, statevoiscell);
            }

            const Rd Mr(xr[node] - Cent);

            const double qtt_limitor_density = rhol / (rhol + dot(grad_rhol, Mr));

            const Rd& High_ordervelocity       = qtt_limitor_density * (grad_ul * Mr);
            const double High_orderinternalnrj = qtt_limitor_density * dot(grad_internal_nrj, Mr);

            const double valLineO2node = internal_nrj + High_orderinternalnrj -
                                         .5 * dot(High_ordervelocity, High_ordervelocity);   // Partie Dukowicz

            minval = std::min(minval, valLineO2node);
            maxval = std::max(maxval, valLineO2node);
          }

          for (size_t l = 0; l < cell_to_face.size(); ++l) {
            const FaceId& face = cell_to_face[l];

            const Rd Mb(xface[face] - Cent);

            const double qtt_limitor_density = rhol / (rhol + dot(grad_rhol, Mb));

            const Rd& High_ordervelocity       = qtt_limitor_density * (grad_ul * Mb);
            const double High_orderinternalnrj = qtt_limitor_density * dot(grad_internal_nrj, Mb);

            const double valLineO2face = internal_nrj + High_orderinternalnrj -
                                         .5 * dot(High_ordervelocity, High_ordervelocity);   // Partie Dukowicz

            minval = std::min(minval, valLineO2face);
            maxval = std::max(maxval, valLineO2face);
          }

          for (size_t l = 0; l < cell_to_edge.size(); ++l) {
            const EdgeId& edge = cell_to_edge[l];

            const Rd Me(xedge[edge] - Cent);

            const double qtt_limitor_density = rhol / (rhol + dot(grad_rhol, Me));

            const Rd& High_ordervelocity       = qtt_limitor_density * (grad_ul * Me);
            const double High_orderinternalnrj = qtt_limitor_density * dot(grad_internal_nrj, Me);

            const double valLineO2edge = internal_nrj + High_orderinternalnrj -
                                         .5 * dot(High_ordervelocity, High_ordervelocity);   // Partie Dukowicz

            minval = std::min(minval, valLineO2edge);
            maxval = std::max(maxval, valLineO2edge);
          }

          static const double epsZer0 = 1e-15;

          // on applique le traitement idem de l'ordre 2..
          double coef1 = 0, coef2 = 0;
          if (enableWeakBoundPositivityOnly) {
            coef1                  = 1;
            const double minGlobal = 1e-12;

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minGlobal - State) / ((minval - State));

          } else {
            if (std::fabs(maxval - State) <= epsZer0)   // epsIsEqualAbs())
              coef1 = 1.;
            else
              coef1 = (maxvalvois - State) / ((maxval - State));

            if (std::fabs(minval - State) <= epsZer0)   // epsIsEqualAbs())
              coef2 = 1.;
            else
              coef2 = (minvalvois - State) / ((minval - State));
          }
          const double alfa = std::max(0., std::min(1., std::min(coef1, coef2)));

          Limitor_eps[j] = alfa;
        });
    }
    synchronize(Limitor_eps);
  }
};
#endif
